(*	:Copyright:  1994 by Stephan Kaufmann

	Permission is hereby granted to make copies of this file for any 
	purpose other than direct profit, or as part of a commercial product,
	provided this copyright notice is left intact.
	Sale is prohibited.
*)

(*	:Name: SKPackages`Linearization` *)

(*	:Author:
	Stephan Kaufmann
	Institute of Mechanics
	Swiss Federal Institute of Technology
	ETH Zentrum, HG F38.4 
	CH-8092 Zurich
	Switzerland
	kaufmann@ifm.mavt.ethz.ch
*)

(*	:Package Version: 2.2 *)

(*	:Mathematica Version: 2.2 *)

(*	:History:
	Changes in Version 2.2 (August 93):
	- Now returns a single expression when given a single expression
	- Works correctly for equations which evaluate to True or False when
	  linearized
*)

(*	:Keywords:
	Linearization, Jacobi matrix, Linearize
*)

(*	:Requirements: none *)

(*	:Limitations: none *)

(*	:Discussion:

	JacobiMatrix[functions, variables, atPoint] calculates
	(at a given point x0 (atPoint)) the Jacobi matrix of a mapping
	f: Rn -> Rm
	with coordinate functions fi (i = 1, ..., m) (functions) in the
	variables xj (j = 1, ..., n) (variables), i.e.: the matrix:
	(dfi/dxj) | xj = x0j (i = 1, ..., m; j = 1, ..., n).
	
	Linearize[functions, variables, atPoint] gives the linearization
	of one or more functions or equations in the given variables,
	at the given point.
*)

(*	:Summary:
	Calculation of the Jacobi matrix and the linearization of
	functions and equations.
*)

If[$VersionNumber < 2.0, Print[
  "This package requires Mathematica version 2.0 or higher."]]
  
BeginPackage["SKPackages`Linearization`"];
JacobiMatrix::usage = 
"JacobiMatrix[functions, variables, atPoint] returns the Jacobi
Matrix of the functions with respect to the variables at the point
atPoint.";

Linearize::usage = 
"Linearize[functions, variables, atPoint] calculates the linearization
of the given functions or equations in the variable at the point atPoint.
The length of variables and atPoint should be equal.";

Linearize::mix =
"You can not mix equations with functions.";

Begin["`Private`"];

JacobiMatrix[functions_, variables_, atPoint_] :=
  (Outer[D, Flatten[{functions}], Flatten[{variables}]] /. 
    Thread[variables -> atPoint]) /;
      Length[variables] == Length[atPoint]
	  
Linearize[functions_List, variables_, atPoint_] :=
  Module[
    {
      subst = Thread[variables -> atPoint],
      varList = Flatten[{variables}],
      calculateIt, eq
    },
    calculateIt = ((# /. subst) +
      (Outer[D, #, varList] /. subst) . varList)&;
    Which[
      FreeQ[functions, Equal], (* only functions *)
        calculateIt[functions],
      And @@ (Head[#] === Equal& /@ functions), (* only equations *)
        (* prevent evaluation of Equal *)
        Thread[calculateIt /@
          Thread[functions /. Equal -> eq, eq]] /. eq -> Equal,
      True, (* else *)
        Message[Linearize::mix]
    ]
  ] /; Length[variables] == Length[atPoint]

Linearize[functions_, variables_, atPoint_] :=
  Linearize[{functions}, variables, atPoint][[1]] /;
    Length[variables] == Length[atPoint]
	
End[ ];

EndPackage[ ];
