(*	:Copyright:  1994 by Stephan Kaufmann

	Permission is hereby granted to make copies of this file for any 
	purpose other than direct profit, or as part of a commercial product,
	provided this copyright notice is left intact.
	Sale is prohibited.
*)

(*	:Name: SKPackages`RSA` *)

(*	:Author:
	Stephan Kaufmann
	Institute of Mechanics
	Swiss Federal Institute of Technology
	ETH Zentrum, HG F38.4 
	CH-8092 Zurich
	Switzerland
	kaufmann@ifm.mavt.ethz.ch
*)

(*	:Package Version: 2.2 *)

(*	:Mathematica Version: 2.2 *)

(*	:History:
	Changes in Version 2.2 (August 93):
	None
*)

(*	:Keywords:
	RSA, public key, cryptosystem
*)

(*	:Requirements: none *)

(*	:Limitations: none *)

(*	:Discussion:
	See e.g. Akritas for a short discussion of the RSA cryptosystem or
	the original paper cited below.
	This package contains tools to work with RSA.
	
	Akritas, A. G.: Elements of Computer Algebra with Applications.
	Wiley, New York, 1989.

	Rivest, R. L., A. Shamir and L. Adleman: A method for obtaining
	digital signatures and public key cryptosystems.
	Communications of the ACM 21, 120-126, 1978.
	
*)

(*	:Summary:
	Demonstration of the RSA public key encryption system.
*)

If[$VersionNumber < 2.0, Print[
  "This package requires Mathematica version 2.0 or higher."]]
  
BeginPackage["SKPackages`RSA`"];

RSAPrime::usage =
"RSAPrime[digits] returns a prime suitable for RSA with the given
number of digits.";

GenerateE::usage =
"GenerateE[p, q] generates a random encoding exponent e which is
relatively prime to (p-1)(q-1).";

Off[General::spell1];

GenerateD::usage =
"GenerateD[e, p, q] returns the multiplicative inverse of
e mod((p-1)(q-1)).";

On[General::spell1];

StringToList::usage =
"StringToList[text, n] converts a given string to block of numbers
smaller than n.";

ListToString::usage =
"ListToString[l] converts a list of numbers (resulting from
StringToList) to the original text string.";

RSAEncodeNumber::usage =
"RSAEncodeNumber[num, e, n] encodes the given number num using
RSA with exponent e and modulus n.";

RSADecodeNumber::usage =
"RSADecodeNumber[num, d, n] decodes the given number num using RSA
with exponent d and modulus n.";

RSAEncode::usage =
"RSAEncode[text, e, n] encodes the given text by converting it to a
list of numbers and using RSA with exponent e and modulus n.";

RSADecode::usage =
"RSADecode[l, d, n] decodes the given list of numbers l by using RSA
with exponent d and modulus n and converting the result back to the
original text string.";

Begin["`Private`"];

GoodRSAPrimeQ[n_Integer, l_Integer] :=
  PrimeQ[n] &&
  Module[{d = 1, m = (n-1)/2},
    While[d <= l && !PrimeQ[m/d], d++];
    d <= l
  ]
  
RSAPrime[digits_Integer] :=
  Module[{l, cand},
    cand = Random[Integer, {10^(digits-1), 10^digits-1}];
    l = 10^Floor[Log[10., cand]/30];
    If[EvenQ[cand], cand++];
    While[!GoodRSAPrimeQ[cand, l], cand += 2];
    cand
  ]
  
GenerateE[p_Integer, q_Integer] :=
  Module[{res, n = p q, phi = (p-1)(q-1)},
    res = Random[Integer, {p, n}];
    While[GCD[res, phi] != 1,
      res = Random[Integer, {p, n}]];
    res
  ]
  
GenerateD[e_Integer, p_Integer, q_Integer] :=
  PowerMod[e, -1, (p-1)(q-1)]
  
RSAEncodeNumber[num_Integer, e_Integer, n_Integer] :=
  PowerMod[num, e, n] /; num < n
  
RSADecodeNumber[num_Integer, d_Integer, n_Integer] :=
  PowerMod[num, d, n]
  
ConvertString[str_String] :=
  Fold[Plus[256 #1, #2]&, 0, ToCharacterCode[str]]
  
StringToList[text_String, n_Integer] :=
  Module[
    { 
      blockLength = Floor[N[Log[256, n]]],
      strLength = StringLength[text]
    },    
    ConvertString /@ Table[StringTake[text, 
        {i, Min[strLength, i + blockLength - 1]}],
      {i, 1, strLength, blockLength}]
  ] /; n > 256
  
MakeList[0] = {};

MakeList[num_Integer] :=
  Append[MakeList[Quotient[num, 256]], Mod[num, 256]]
  
ConvertNumber[num_Integer] := FromCharacterCode /@ MakeList[num]

ListToString[l_List] :=
  StringJoin[ConvertNumber /@ l]
  
RSAEncode[text_String, e_Integer, n_Integer] :=
  RSAEncodeNumber[#, e, n]& /@
    StringToList[text, n] /; n > 256
	
RSADecode[l_List, d_Integer, n_Integer] :=
  ListToString[RSADecodeNumber[#, d, n]& /@ l]
  
End[ ];

EndPackage[ ];
